#!/usr/bin/env python
# -*- coding: utf-8 -*-

# |>
# | DATA FORMATS ANALYZER 1-1-1 - DFA PY [http://www.universalchemicalmarkup.org]
# | Updated: 2015-08-17
# | License: GNU GPL 3 [http://www.gnu.org/licenses/gpl-3.0.html]
# | Author: Jan Mokrý (Main corresponding author:
# |                    Contact information is at the project website.)
# |>

import sys

import DfaOutput
import DfaConverter
import DfaExtractor
import DfaPreviewer

# DFA PY - INIT SERVICE
class InitService(object):
    '''
    Provides methods to initialize other modules.

    **Parameters:**

    - _`verbosity`: The *string* specifying verbosity. Possible values are:

      - 'IWE' - Output INFO, WARNING and ERROR messages.
      - 'WE' - Output WARNING and ERROR messages.
      - 'E' - Output ERROR messages.

    - _`mode`: The *string* specifying the mode of operation. Possible values are:

      - 'X' - Extract to XHTML and XML.
      - 'CD' - Convert DSV to XML.
      - 'CX' - Convert XML to DSV and TEX.
      - 'P' - Preview updated XML by updating XHTML and XML.

    - filepath: The *string* specifying the full path to the given XSD schema, which will be parsed.
    - _`version`: The *string* specifying the output version.
    - _`refName`: The *string* specifying the reference name.
    - _`refVersion`: The *string* specifying the reference version.
    - rootElementName: The *string* specifying the root element name for which the tree structure should be generated.

    Note that the parameters version_, refName_ and refVersion_ are used for generating filenames to save or load various output
    files.

    The values of following class variables can be adjusted in the source code to fit the needs of the specific analysis:
    *dsvDelimiter*, *outputDocumentation*, *outputStandard*, *standardStringParts*, *surroundingChars* and *adjustableFilenames*.
    '''

    def __init__(self, verbosity='IWE', mode='X', filepath=None, version=None, refName=None, refVersion=None,
                 rootElementName=None):
        '''
        Initializes class variables.
        '''
        self.outputService = DfaOutput.OutputService(verbosity=verbosity, version=version, refName=refName, refVersion=refVersion)
        self.argVerbosity = verbosity
        self.argMode = mode
        self.argFilepath = filepath
        self.argVersion = version
        self.argRefName = refName
        self.argRefVersion = refVersion
        self.argRootElementName = rootElementName
        # Specify the delimiter for DSV files. Use comma (',') or tab ('\t') for compatibility.
        # Adjust this to fit the needs of the specific analysis.
        self.dsvDelimiter = ','
        # Specify whether the description should be generated for each attribute, element and type from the parsed XSD schema.
        # Note that the description is extracted from documentation annotations in the parsed XSD schema. Use true (True) only
        # when the given XSD schema contains documentation annotations, otherwise use false (False).
        # Use none (None) to be asked each time this information is needed. Adjust this to fit the needs of the specific analysis.
        self.outputDocumentation = None
        # Specify whether the link to the standard should be generated for each attribute, element and type
        # from the parsed XSD schema. Note that the standard is either the official documentation or the schema in cases when
        # the documentation is not available in such a detailed form or uses some other data format than HTML or XHTML.
        # Use true (True) only when the given XSD schema is supported, otherwise use false (False) or add the support for
        # the XSD schema by providing corresponding standard string parts.
        # Use none (None) to be asked each time this information is needed. Adjust this to fit the needs of the specific analysis.
        self.outputStandard = None
        # The standard string parts used to generate the link to the standard for each attribute, element and type
        # from the parsed XSD schema. Adjust this to fit the needs of the specific analysis.
        self.standardStringParts = {
            # To add standard string parts for a new format adjust and uncomment the following entry:
            #'EXAMPLE-FORMAT--1-1':('http://www.example.com/format/', '(T)/(N).html'),
            # where the 'http://www.example.com/format/' is the fixed part of the link to the standard,
            # and '(T)/(N).html' is the variable part of the link to the standard.
            # In the variable part the '(T)' value will be replaced by 'attribute', 'element' or 'type',
            # and '(N)' value will be replaced by the name of the current attribute, element or type.
            # This makes it possible to output direct links to the standard for each attribute, element or type
            # from the XML formats that have sufficiently structured official documentations.
            'CDXML--4-8':('http://www.cambridgesoft.com/xml/cdxml.dtd', ''),
            'CML--3':('http://xml-cml.org/schema/schema3/schema.xsd', ''),
            'MMDBCHEMICALGRAPHXML--20141117':(
                'http://www.ncbi.nlm.nih.gov/data_specs/ver/20141117/schema/MMDB_Chemical_graph_20141117.mod.xsd', ''),
            'MMDBSTRUCTURALMODELXML--20141117':(
                'http://www.ncbi.nlm.nih.gov/data_specs/ver/20141117/schema/MMDB_Structural_model_20141117.mod.xsd', ''),
            'PCSUBSTANCEXML--20141117':(
                'http://www.ncbi.nlm.nih.gov/data_specs/ver/20141117/schema/NCBI_PCSubstance_20141117.mod.xsd', ''),
            'PDBML--4-0':('http://pdbml.pdb.org/schema/pdbx-v40.xsd', ''),
            'PDBML--4-2':('http://pdbml.pdb.org/schema/pdbx-v42.xsd', ''),
            'UCM--1-1-1':('http://www.universalchemicalmarkup.org', ''),
            'XDFILE--20110328':('http://accelrys.com/products/informatics/cheminformatics/ctfile-formats/no-fee.php', '')}
        # Chars that can surround attribute and element names in descriptions. It is used for marking the occurrences of attribute
        # and element names to generate cross-references in descriptions. Adjust this to fit the needs of the specific analysis.
        self.surroundingChars = [
            (' ', ' '), (' ', '<'), ('>', ' '), ('&lt;', ' '), ('&lt;', '&gt;'), ('&quot;', '&quot;'), (' ', '=&quot;'),
            (' ', ':'), (':', ' '), (' ', '.'), ('.', ' '), (' ', ','), (',', ' ')]
        # Filenames of manualy or semi-automatically created and edited files.
        # Adjust this to fit the needs of the specific analysis.
        self.adjustableFilenames = [
            'Ov--AnalyzedFormats--' + self.argVersion,
            'Ov--FullyAnalyzedFormats--' + self.argVersion,
            'Ov--FullyAnalyzedFormatsAssessments--' + self.argVersion,
            'Ov--ExcludedFormats--' + self.argVersion,
            'OvRef--ChosenConcepts--' + self.argVersion,
            'OvRef--TargetConcepts--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Attributes--ExtractedInfoAndDescription--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Elements--ExtractedInfoAndDescription--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Types--ExtractedInfoAndDescription--' + self.argVersion]
        # Filenames of automatically generated files (those can be automatically or manualy edited and updated).
        self.automaticFilenames = [
            'OvRef--MergedConcepts--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Attributes--ExtractedInfo--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Elements--ExtractedInfo--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Types--ExtractedInfo--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Attributes--ExtractedDescription--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Elements--ExtractedDescription--' + self.argVersion,
            'Ref--' + self.argRefName + '--' + self.argRefVersion + '--Types--ExtractedDescription--' + self.argVersion]
        self.currentFilenames = self.adjustableFilenames + self.automaticFilenames
        self.currentExtension = '.xml'

    def initJob(self):
        '''
        Initializes and executes appropriate modules for a job specified by the mode_.
        '''
        if self.argMode == 'X':
            self._executeExtractor()
        elif self.argMode == 'CD':
            self._setCurrentExtension()
            self._executeConverter()
        elif self.argMode == 'CX':
            self._executeConverter()
        elif self.argMode == 'P':
            self._executePreviewer()

    def _executeConverter(self):
        '''
        Initializes and executes Converter.
        '''
        self.outputService.printMessage(message='INFO: Converter module will be executed.')
        convertService = DfaConverter.ConvertService(verbosity=self.argVerbosity, dsvDelimiter=self.dsvDelimiter,
                                                  refName=self.argRefName, refVersion=self.argRefVersion)

        for currentFilename in self.currentFilenames:
            convertService.convertData(filepath=currentFilename + self.currentExtension)

    def _executeExtractor(self):
        '''
        Initializes and executes Extractor.
        '''
        self.outputService.printMessage(message='INFO: Extractor module will be executed.')

        if (self.outputDocumentation is None) or (self.outputStandard is None):
            print 'Specify what should be outputed for each extracted attribute, element and type:'

            if self.outputDocumentation is None: self.outputDocumentation = raw_input(
                'Output description [ENTER = No / Y+ENTER = Yes]: ')

            if self.outputStandard is None: self.outputStandard = raw_input(
                'Output link to standard [ENTER = No / Y+ENTER = Yes]: ')

        extractService = DfaExtractor.ExtractService(verbosity=self.argVerbosity, filepath=self.argFilepath)
        extractedAttributes, extractedElements, extractedTypes = extractService.extractData(
            extractDescriptions=self.outputDocumentation)
        self.outputService.outputData(rootElementName=self.argRootElementName,
            extractedAttributes=extractedAttributes, extractedElements=extractedElements, extractedTypes=extractedTypes,
            surroundingChars=self.surroundingChars, standardStringParts=self.standardStringParts,
            outputStandard=self.outputStandard)

    def _executePreviewer(self):
        '''
        Initializes and executes Previewer.
        '''
        self.outputService.printMessage(message='INFO: Previewer module will be executed.')
        previewService = DfaPreviewer.PreviewService(verbosity=self.argVerbosity, filepath=self.argFilepath, version=self.argVersion,
                                                  refName=self.argRefName, refVersion=self.argRefVersion)
        previewService.previewData()

    def _setCurrentExtension(self):
        '''
        Sets the current extension.
        '''
        if self.dsvDelimiter == ',':
            self.currentExtension = '.csv'
        elif self.dsvDelimiter == '\t':
            self.currentExtension = '.tsv'

if __name__ == '__main__':
    if len(sys.argv) != 8:
        print 'DATA FORMATS ANALYZER - QUICK HELP'
        print 'Required parameters:'
        print ' (1)-Verbosity'
        print ' (2)-Mode'
        print ' (3)-Schema filepath'
        print ' (4)-Output version'
        print ' (5)-Ref name'
        print ' (6)-Ref version'
        print ' (7)-Root element name'
        print 'Possible verbosity values:'
        print ' "IWE" - Output INFO, WARNING and ERROR messages.'
        print ' "WE" - Output WARNING and ERROR messages.'
        print ' "E" - Output ERROR messages.'
        print 'Possible mode values:'
        print ' "X" - Extract to XHTML and XML.'
        print ' "CD" - Convert DSV to XML.'
        print ' "CX" - Convert XML to DSV and TEX.'
        print ' "P" - Preview updated XML by updating XHTML and XML.'
    else:
        print 'DATA FORMATS ANALYZER - RUNNING'
        initService = InitService(verbosity=sys.argv[1], mode=sys.argv[2], filepath=sys.argv[3], version=sys.argv[4],
                                  refName=sys.argv[5], refVersion=sys.argv[6], rootElementName=sys.argv[7])
        initService.initJob()
        print 'DATA FORMATS ANALYZER - FINISHED'
