#!/usr/bin/env python
# -*- coding: utf-8 -*-

# |>
# | DATA FORMATS ANALYZER 1-1-1 - DFA EXTRACTOR PY [http://www.universalchemicalmarkup.org]
# | Updated: 2014-08-12
# | License: GNU GPL 3 [http://www.gnu.org/licenses/gpl-3.0.html]
# | Author: Jan Mokrý (Main corresponding author:
# |                    Contact information is at the project website.)
# |>

import re
import sys

import xml.dom.minidom
import xml.sax.saxutils

import DfaOutput

# DFA EXTRACTOR PY - EXTRACT SERVICE
class ExtractService(object):
    '''
    Provides methods to extract information and descriptions data about attributes, elements and types
    from the given XSD schema.

    Attributes data are stored in the _`extractedAttributes` *dictionary* with the following structure::

        DICTIONARY{
            Attribute name (STRING):DICTIONARY{
                parents (STRING):LIST[Parent name (STRING)],
                description (STRING):LIST[Attribute documentation (STRING)]
            }
        }

    Elements data are stored in the _`extractedElements` *dictionary* with the following structure::

        DICTIONARY{
            Element name (STRING):DICTIONARY{
                attributes (STRING):LIST[Attribute name (STRING)],
                children (STRING):LIST[Child name (STRING)],
                parents (STRING):LIST[Parent name (STRING)],
                description (STRING):LIST[Element documentation (STRING)]
            }
        }

    Types data are stored in the _`extractedTypes` *dictionary* with the following structure::

        DICTIONARY{
            Type name (STRING):DICTIONARY{
                attributes (STRING):LIST[Attribute name (STRING)],
                children (STRING):LIST[Child name (STRING)],

                usages (STRING):DICTIONARY{
                    attributes (STRING):LIST[Attribute name (STRING)],
                    elements (STRING):LIST[Element name (STRING)]
                },

                description (STRING):LIST[Type documentation (STRING)]
            }
        }

    Substitution groups data are stored in the _`substitutionGroups` *dictionary* with the following structure::

        DICTIONARY{
            Head element name (STRING):LIST[Element name (STRING)]
        }

    **Parameters:**

    - verbosity: The *string* specifying verbosity. See details in :ref:`InitService`.
    - filepath: The *string* specifying the full path to the given XSD schema, which will be parsed.
    '''

    def __init__(self, verbosity='IWE', filepath=None):
        '''
        Initializes class variables and parses the given XSD schema.
        '''
        self.outputService = DfaOutput.OutputService(verbosity=verbosity)

        try:
            self.currentSchema = xml.dom.minidom.parse(filepath)
            self.currentAttributeName = ''
            self.currentElementName = ''
            self.currentTypeName = ''
            self.extractedAttributes = {}
            self.extractedElements = {}
            self.extractedTypes = {}
            self.substitutionGroups = {}
        except IOError:
            self.outputService.printMessage(message='ERROR: File not found - "' + filepath + '"!')

    def extractData(self, extractDescriptions=False):
        '''
        Extracts information and if requested also descriptions data about attributes, elements and types from the given XSD schema.

        **Parameters:**

        - _`extractDescriptions`: The *boolean* specifying whether descriptions data should be extracted from the documentation
          annotations in the given XSD schema. Use true (``True``) only when the XSD schema contains documentation annotations,
          otherwise use false (``False``).

        **Returns:** The *tuple* containing extractedAttributes_, extractedElements_ and extractedTypes_ *dictionaries*.
        It has the following structure: ``TUPLE(3*DICTIONARY)``.
        '''
        self.outputService.printMessage(message='INFO: Extracting attributes and elements.')

        for e in self.currentSchema.getElementsByTagNameNS('*', 'element'):
            if e.hasAttribute('name'):
                self.currentElementName = e.getAttribute('name')

                if e.hasAttribute('substitutionGroup'): self._extractAndSaveSubstitutionGroup(e)

                self.currentElementName = e.getAttribute('name')
                self._extractAttributeNames(e, extractDescriptions)
                self._extractElementNames(e)

                if extractDescriptions: self._saveElementDocumentation(e)

        self.outputService.printMessage(message='INFO: Attributes and elements extracted.')
        self.currentElementName = ''
        self.outputService.printMessage(message='INFO: Extracting attributes and types.')

        for e in self.currentSchema.getElementsByTagNameNS('*', 'complexType'):
            if e.hasAttribute('name'):
                self.currentTypeName = 'C-' + e.getAttribute('name')
                self._extractAttributeNames(e, extractDescriptions)
                self._extractElementNames(e)
                self._saveTypeName(self.currentTypeName, extractDescriptions)

                if extractDescriptions: self._saveTypeDocumentation(e)

        for e in self.currentSchema.getElementsByTagNameNS('*', 'simpleType'):
            if e.hasAttribute('name'):
                self.currentTypeName = 'S-' + e.getAttribute('name')
                self._saveTypeName(self.currentTypeName, extractDescriptions)

                if extractDescriptions: self._saveTypeDocumentation(e)

        self.outputService.printMessage(message='INFO: Attributes and types extracted.')
        self._processSubstitutionGroups()
        self.outputService.printMessage(message='INFO: Substitution groups for elements processed.')
        self._processTypeUsages()
        self.outputService.printMessage(message='INFO: Types for elements and attributes processed.')
        return (self.extractedAttributes, self.extractedElements, self.extractedTypes)

    def _extractAttributeNames(self, element, extractDescriptions):
        '''
        Extracts attribute names from a given ``<*:element>`` element in the XSD schema.

        **Parameters:**

        - element: The *xml.dom.minidom.Element* with the current ``<*:element>`` element, which will be used to extract
          attribute names.
        - extractDescriptions_: The *boolean* specifying whether descriptions data should be extracted.
        '''
        for eAttribute in element.getElementsByTagNameNS('*', 'attribute'):
            if eAttribute.hasAttribute('name'):
                self.currentAttributeName = eAttribute.getAttribute('name')
                self._saveAttributeName(self.currentAttributeName)

                if extractDescriptions: self._saveAttributeDocumentation(eAttribute)
            elif eAttribute.hasAttribute('ref'):
                attributeName = eAttribute.getAttribute('ref').split(':')[-1]

                for eAttribute in self.currentSchema.getElementsByTagNameNS('*', 'attribute'):
                    if eAttribute.hasAttribute('name'):
                        if eAttribute.getAttribute('name') == attributeName:
                            self._extractAttributeNames(eAttribute, extractDescriptions)

        for eAttributeGroup in element.getElementsByTagNameNS('*', 'attributeGroup'):
            if eAttributeGroup.hasAttribute('name'): self._extractAttributeNames(eAttributeGroup, extractDescriptions)
            elif eAttributeGroup.hasAttribute('ref'):
                attributeGroupName = eAttributeGroup.getAttribute('ref').split(':')[-1]

                for eAttributeGroup in self.currentSchema.getElementsByTagNameNS('*', 'attributeGroup'):
                    if eAttributeGroup.hasAttribute('name'):
                        if eAttributeGroup.getAttribute('name') == attributeGroupName:
                            self._extractAttributeNames(eAttributeGroup, extractDescriptions)

    def _saveAttributeName(self, attributeName):
        '''
        Saves the attribute name in the extractedAttributes_ *dictionary* and in the attributes lists
        of extractedElements_ and extractedTypes_ *dictionaries*.

        **Parameters:**

        - attributeName: The *string* specifying the current attribute name, which will be saved.
        '''
        # All possibilities could happen depending on order in which save*Name and save*Documentation methods are called.
        if attributeName not in self.extractedAttributes: self.extractedAttributes[attributeName] = {}

        if self.currentElementName == '':
            if self.currentTypeName not in self.extractedTypes:
                self.extractedTypes[self.currentTypeName] = {'attributes':[attributeName]}
            elif 'attributes' not in self.extractedTypes[self.currentTypeName]:
                self.extractedTypes[self.currentTypeName]['attributes'] = [attributeName]
            elif attributeName not in self.extractedTypes[self.currentTypeName]['attributes']:
                self.extractedTypes[self.currentTypeName]['attributes'].append(attributeName)
        else:
            if 'parents' not in self.extractedAttributes[attributeName]:
                self.extractedAttributes[attributeName]['parents'] = [self.currentElementName]
            elif self.currentElementName not in self.extractedAttributes[attributeName]['parents']:
                self.extractedAttributes[attributeName]['parents'].append(self.currentElementName)

            if self.currentElementName not in self.extractedElements:
                self.extractedElements[self.currentElementName] = {'attributes':[attributeName]}
            elif 'attributes' not in self.extractedElements[self.currentElementName]:
                self.extractedElements[self.currentElementName]['attributes'] = [attributeName]
            elif attributeName not in self.extractedElements[self.currentElementName]['attributes']:
                self.extractedElements[self.currentElementName]['attributes'].append(attributeName)

    def _saveAttributeDocumentation(self, element):
        '''
        Saves the attribute documentation in the extractedAttributes_ *dictionary*.

        **Parameters:**

        - element: The *xml.dom.minidom.Element* with the current ``<*:attribute>`` element, which will be used
          to extract the attribute documentation.
        '''
        for aDocumentation in element.getElementsByTagNameNS('*', 'documentation'):
            if aDocumentation.parentNode.parentNode == element:
                attributeDocumentation = self._cleanupDocumentation(1, aDocumentation)

                # All possibilities could happen when order
                # in which saveAttributeName and saveAttributeDocumentation methods are called is changed.
                if self.currentAttributeName not in self.extractedAttributes:
                    self.extractedAttributes[self.currentAttributeName] = {'description':[attributeDocumentation]}
                elif 'description' not in self.extractedAttributes[self.currentAttributeName]:
                    self.extractedAttributes[self.currentAttributeName]['description'] = [attributeDocumentation]
                elif attributeDocumentation not in self.extractedAttributes[self.currentAttributeName]['description']:
                    self.extractedAttributes[self.currentAttributeName]['description'].append(attributeDocumentation)

    def _extractElementNames(self, element):
        '''
        Extracts element names from a given ``<*:element>`` element in the XSD schema.

        **Parameters:**

        - element: The *xml.dom.minidom.Element* with the current ``<*:element>`` element, which will be used to extract
          element names.
        '''
        if self.currentElementName and (self.currentElementName not in self.extractedElements):
            self.extractedElements[self.currentElementName] = {}

        for eElement in element.getElementsByTagNameNS('*', 'element'):
            if eElement.hasAttribute('name'): self._saveElementName(eElement.getAttribute('name'))
            elif eElement.hasAttribute('ref'): self._saveElementName(eElement.getAttribute('ref').split(':')[-1])

    def _saveElementName(self, elementName):
        '''
        Saves the element name in the extractedElements_ *dictionary*, children and parents lists
        of extractedElements_ *dictionary* and in the children lists of extractedTypes_ *dictionary*.

        **Parameters:**

        - elementName: The *string* specifying the current element name, which will be saved.
        '''
        # All possibilities could happen depending on order in which save*Name and save*Documentation methods are called.
        if elementName not in self.extractedElements: self.extractedElements[elementName] = {}

        if self.currentElementName == '':
            if self.currentTypeName not in self.extractedTypes:
                self.extractedTypes[self.currentTypeName] = {'children':[elementName]}
            elif 'children' not in self.extractedTypes[self.currentTypeName]:
                self.extractedTypes[self.currentTypeName]['children'] = [elementName]
            elif elementName not in self.extractedTypes[self.currentTypeName]['children']:
                self.extractedTypes[self.currentTypeName]['children'].append(elementName)
        else:
            if self.currentElementName not in self.extractedElements:
                self.extractedElements[self.currentElementName] = {'children':[elementName]}
            elif 'children' not in self.extractedElements[self.currentElementName]:
                self.extractedElements[self.currentElementName]['children'] = [elementName]
            elif elementName not in self.extractedElements[self.currentElementName]['children']:
                self.extractedElements[self.currentElementName]['children'].append(elementName)

            if 'parents' not in self.extractedElements[elementName]:
                self.extractedElements[elementName]['parents'] = [self.currentElementName]
            elif self.currentElementName not in self.extractedElements[elementName]['parents']:
                self.extractedElements[elementName]['parents'].append(self.currentElementName)

    def _saveElementDocumentation(self, element):
        '''
        Saves the element documentation in the extractedElements_ *dictionary*.

        **Parameters:**

        - element: The *xml.dom.minidom.Element* with the current ``<*:element>`` element, which will be used to extract
          the element documentation.
        '''
        for eDocumentation in element.getElementsByTagNameNS('*', 'documentation'):
            if eDocumentation.parentNode.parentNode == element:
                elementDocumentation = self._cleanupDocumentation(1, eDocumentation)

                # All possibilities could happen when order
                # in which saveElementName and saveElementDocumentation methods are called is changed.
                if self.currentElementName not in self.extractedElements:
                    self.extractedElements[self.currentElementName] = {'description':[elementDocumentation]}
                elif 'description' not in self.extractedElements[self.currentElementName]:
                    self.extractedElements[self.currentElementName]['description'] = [elementDocumentation]
                elif elementDocumentation not in self.extractedElements[self.currentElementName]['description']:
                    self.extractedElements[self.currentElementName]['description'].append(elementDocumentation)

    def _saveTypeName(self, typeName, extractDescriptions):
        '''
        Saves the type name in the extractedTypes_ *dictionary*.

        **Parameters:**

        - typeName: The *string* specifying the current type name, which will be saved.
        - extractDescriptions_: The *boolean* specifying whether descriptions data should be extracted.
        '''
        if typeName not in self.extractedTypes: self.extractedTypes[typeName] = {}

        if typeName.startswith('C-'): self._saveTypeUsages(typeName, 'element', extractDescriptions)
        else:
            self._saveTypeUsages(typeName, 'attribute', extractDescriptions)
            self._saveTypeUsages(typeName, 'element', extractDescriptions)

    def _saveTypeUsages(self, typeName, usageName, extractDescriptions):
        '''
        Saves and extracts type usages to the usage lists of extractedTypes_ *dictionary*.

        **Parameters:**

        - typeName: The *string* specifying the current type name, which will be saved.
        - usageName: The *string* specifying the current type usage name, which will be extracted and saved.
        - extractDescriptions_: The *boolean* specifying whether descriptions data should be extracted.
        '''
        typeUsages = []

        for e in self.currentSchema.getElementsByTagNameNS('*', usageName):
            if e.hasAttribute('name') and (typeName[2:] == e.getAttribute('type').split(':')[-1]):
                name = e.getAttribute('name')
                typeUsages.append(name)

                # Save information and if requested also descriptions data about attributes not used in any elements.
                if usageName == 'attribute':
                    if name not in self.extractedAttributes: self.extractedAttributes[name] = {}

                    if extractDescriptions:
                        for aDocumentation in e.getElementsByTagNameNS('*', 'documentation'):
                            if aDocumentation.parentNode.parentNode == e:
                                attributeDocumentation = self._cleanupDocumentation(1, aDocumentation)

                                if 'description' not in self.extractedAttributes[name]:
                                    self.extractedAttributes[name]['description'] = [attributeDocumentation]
                                elif attributeDocumentation not in self.extractedAttributes[name]['description']:
                                    self.extractedAttributes[name]['description'].append(attributeDocumentation)

        if typeUsages != []:
            # Eliminate the duplicated usages of the given type - defining more elements or attributes with the same name.
            typeUsages = list(set(typeUsages))
            usageName = usageName + 's'

            if 'usages' not in self.extractedTypes[typeName]: self.extractedTypes[typeName]['usages'] = {usageName:typeUsages}
            elif usageName not in self.extractedTypes[typeName]['usages']:
                self.extractedTypes[typeName]['usages'][usageName] = typeUsages

    def _saveTypeDocumentation(self, element):
        '''
        Saves the type documentation in the extractedTypes_ *dictionary*.

        **Parameters:**

        - element: The *xml.dom.minidom.Element* with the current ``<*:complexType>`` or ``<*:simpleType>`` element, which will be
          used to extract the type documentation.
        '''
        for tDocumentation in element.getElementsByTagNameNS('*', 'documentation'):
            if tDocumentation.parentNode.parentNode == element:
                typeDocumentation = self._cleanupDocumentation(1, tDocumentation)

                # All possibilities could happen when order
                # in which saveTypeName and saveTypeDocumentation methods are called is changed.
                if self.currentTypeName not in self.extractedTypes:
                    self.extractedTypes[self.currentTypeName] = {'description':[typeDocumentation]}
                elif 'description' not in self.extractedTypes[self.currentTypeName]:
                    self.extractedTypes[self.currentTypeName]['description'] = [typeDocumentation]
                elif typeDocumentation not in self.extractedTypes[self.currentTypeName]['description']:
                    self.extractedTypes[self.currentTypeName]['description'].append(typeDocumentation)

    def _cleanupDocumentation(self, maincall, element):
        '''
        Returns a clean documentation for saving in the extractedAttributes_, extractedElements_ or extractedTypes_ *dictionary*.

        **Parameters:**

        - maincall: The *boolean* specifying whether this method is being called by another method (``True``)
          or recursively (``False``).
        - element: The *xml.dom.minidom.Element* with the current ``<*:documentation>`` element, which will be used
          to extract the cleaned documentation.

        **Returns:** The *string* with a clean documentation from the given ``<*:documentation>`` element.
        '''
        description = ''

        for node in element.childNodes:
            if isinstance(node, xml.dom.minidom.Element):
                nodeName = node.tagName

                if (nodeName == 'pre') or nodeName.endswith(':pre'):
                    # Eliminate the namespace part in the <*:pre> element's name.
                    description += '\n' + node.toxml().replace(
                        '<' + nodeName + '>', '<pre>').replace('</' + nodeName + '>', '</pre>') + '\n'
                elif (nodeName == 'div') or (nodeName == 'p') or nodeName.endswith(':div') or nodeName.endswith(':p'):
                    descriptionPart = self._cleanupDocumentationText(node)

                    # Ensure there is no <pre> element inside before encapsulating the current documentation string part.
                    if '<' not in descriptionPart: description = description.strip() + '\n<p>' + descriptionPart + '</p>'
                    else: description = description.strip() + '\n' + descriptionPart
                else: description += self._cleanupDocumentation(0, node) + ' '
            elif isinstance(node, xml.dom.minidom.Text):
                text = node.toxml()

                if text.startswith('@REPLACE'): description += text + ' '
                else: description += ' '.join(text.split()) + ' '

        # Process instructions found in the current documentation string.
        if description.startswith('@REPLACE') and maincall:
            iPattern = re.compile('@REPLACE\\{(\\[(\\(\\\'.*?\\\',\\\'.*?\\\'\\),?)+\\])\\}')
            iList = eval(xml.sax.saxutils.unescape(iPattern.match(description).group(1)))
            description = iPattern.sub(r'', description)

            for replaced, replacement in iList: description = description.replace(replaced, replacement)

        # Ensure there is no <p> or <pre> element inside before encapsulating the current documentation string.
        if ('<' not in description) and maincall: description = '<p>' + description.strip() + '</p>'

        if maincall: description = description.replace('<p></p>', '')

        return description.strip()

    def _cleanupDocumentationText(self, element):
        '''
        Returns a clean documentation text from the given child element of the ``<*:documentation>`` element.

        **Parameters:**

        - element: The *xml.dom.minidom.Element* with the current child element of the ``<*:documentation>`` element, which will be
          used to extract the cleaned documentation text.

        **Returns:** The *string* with a clean documentation text from the given child element of the ``<*:documentation>`` element.
        '''
        text = ''

        for node in element.childNodes:
            if isinstance(node, xml.dom.minidom.Element):
                nodeName = node.tagName

                if (nodeName == 'pre') or nodeName.endswith(':pre'):
                    # Eliminate the namespace part in the <*:pre> element's name.
                    text += '\n' + node.toxml().replace(
                        '<' + nodeName + '>', '<pre>').replace('</' + nodeName + '>', '</pre>') + '\n'
                else: text += self._cleanupDocumentationText(node) + ' '
            elif isinstance(node, xml.dom.minidom.Text): text += ' '.join(node.toxml().split()) + ' '

        return text.strip()

    def _extractAndSaveSubstitutionGroup(self, element):
        '''
        Extracts and saves a substitution group for the given ``<*:element>`` element in the substitutionGroups_ *dictionary*.

        **Parameters:**

        - element: The *xml.dom.minidom.Element* with the current ``<*:element>`` element, which will be used to extract
          the element name and substitution group.
        '''
        substitutionGroup = element.getAttribute('substitutionGroup')

        if substitutionGroup not in self.substitutionGroups: self.substitutionGroups[substitutionGroup] = [self.currentElementName]
        elif self.currentElementName not in self.substitutionGroups[substitutionGroup]:
            self.substitutionGroups[substitutionGroup].append(self.currentElementName)

    def _processSubstitutionGroups(self):
        '''
        Processes the substitutionGroups_ *dictionary* to update the extractedElements_ *dictionary* to take substitution groups
        into account.
        '''
        if self.substitutionGroups:
            for substitutionGroup in self.substitutionGroups:
                for e in self.currentSchema.getElementsByTagNameNS('*', 'element'):
                    if e.hasAttribute('name') and e.hasAttribute('abstract'):
                        if (e.getAttribute('name') == substitutionGroup) and (e.getAttribute('abstract') == 'true'):
                            self.substitutionGroups[substitutionGroup].insert(0, False)
                            break

            for elementName in self.extractedElements:
                if 'parents' in self.extractedElements[elementName]: eParents = self.extractedElements[elementName]['parents']
                else: eParents = []

                for substitutionGroup in self.substitutionGroups:
                    if elementName in self.substitutionGroups[substitutionGroup]:
                        if 'parents' in self.extractedElements[substitutionGroup]:
                            eParents = list(set(eParents + self.extractedElements[substitutionGroup]['parents']))

                if eParents: self.extractedElements[elementName]['parents'] = eParents

            for elementName in self.extractedElements:
                if 'children' in self.extractedElements[elementName]: eChildren = self.extractedElements[elementName]['children']
                else: eChildren = []

                if eChildren:
                    for substitutionGroup in self.substitutionGroups:
                        if substitutionGroup in eChildren:
                            if not self.substitutionGroups[substitutionGroup][0]:
                                eChildren.remove(substitutionGroup)
                                eChildren = list(set(eChildren + self.substitutionGroups[substitutionGroup][1:]))
                            else: eChildren = list(set(eChildren + self.substitutionGroups[substitutionGroup]))

                            self.extractedElements[elementName]['children'] = eChildren

            for substitutionGroup in self.substitutionGroups:
                if not self.substitutionGroups[substitutionGroup][0]:
                    del self.extractedElements[substitutionGroup]

    def _processTypeUsages(self):
        '''
        Processes the extractedTypes_ *dictionary* to update the extractedElements_ and extractedAttributes_ *dictionaries* to take
        types into account.
        '''
        if self.extractedTypes:
            for typeName in self.extractedTypes:
                if 'usages' in self.extractedTypes[typeName]:
                    if 'elements' in self.extractedTypes[typeName]['usages']:
                        tUsagesElements = self.extractedTypes[typeName]['usages']['elements']

                        if 'attributes' in self.extractedTypes[typeName]:
                            tAttributes = self.extractedTypes[typeName]['attributes']

                            for elementName in tUsagesElements:
                                if 'attributes' in self.extractedElements[elementName]:
                                    eAttributes = list(set(self.extractedElements[elementName]['attributes'] + tAttributes))
                                else: eAttributes = tAttributes

                                self.extractedElements[elementName]['attributes'] = eAttributes

                            for attributeName in tAttributes:
                                if 'parents' in self.extractedAttributes[attributeName]:
                                    aParents = list(set(self.extractedAttributes[attributeName]['parents'] + tUsagesElements))
                                else: aParents = tUsagesElements

                                self.extractedAttributes[attributeName]['parents'] = aParents

                        if 'children' in self.extractedTypes[typeName]:
                            tChildren = self.extractedTypes[typeName]['children']

                            for elementName in tUsagesElements:
                                if 'children' in self.extractedElements[elementName]:
                                    eChildren = list(set(self.extractedElements[elementName]['children'] + tChildren))
                                else: eChildren = tChildren

                                self.extractedElements[elementName]['children'] = eChildren

                            for elementName in tChildren:
                                if 'parents' in self.extractedElements[elementName]:
                                    eParents = list(set(self.extractedElements[elementName]['parents'] + tUsagesElements))
                                else: eParents = tUsagesElements

                                self.extractedElements[elementName]['parents'] = eParents
