#!/usr/bin/env python
# -*- coding: utf-8 -*-

# |>
# | DATA FORMATS ANALYZER 1-1-1 - DFA OUTPUT PY [http://www.universalchemicalmarkup.org]
# | Updated: 2015-08-22
# | License: GNU GPL 3 [http://www.gnu.org/licenses/gpl-3.0.html]
# | Author: Jan Mokrý (Main corresponding author:
# |                    Contact information is at the project website.)
# |>

import sys

import datetime

# DFA OUTPUT PY - OUTPUT SERVICE
class OutputService(object):
    '''
    Provides methods to output data from other modules.

    **Parameters:**

    - _`verbosity`: The *string* specifying verbosity. See details in :ref:`InitService`.
    - version: The *string* specifying the output version. See details in :ref:`InitService`.
    - refName: The *string* specifying the reference name. See details in :ref:`InitService`.
    - refVersion: The *string* specifying the reference version. See details in :ref:`InitService`.
    '''

    def __init__(self, verbosity='E', version=None, refName=None, refVersion=None):
        '''
        Initializes class variables.
        '''
        self.currentVerbosity = verbosity
        self.currentVersion = version
        self.refName = refName
        self.refVersion = refVersion

    def printMessage(self, message=None):
        '''
        Prints the message depending on specified verbosity_. In the case of ERROR message it also exits the program.

        **Parameters:**

        - message: The *string* with the message.
        '''
        if isinstance(message, str):
            if message.startswith('INFO: ') or message.startswith('WARNING: '):
                if message[0] in self.currentVerbosity: print message
            else:
                if message.startswith('ERROR: '): sys.exit(message)
                else: sys.exit('ERROR: Message has wrong format - "' + message + '"!')
        else:
            sys.exit('ERROR: Message is not string instance!')

    def outputData(self, rootElementName=None, extractedAttributes=None, extractedElements=None, extractedTypes=None,
                    surroundingChars=None, standardStringParts=None, outputStandard=False):
        '''
        Outputs information and if requested also descriptions data about attributes, elements and types into XHTML and XML files.

        **Parameters:**

        - rootElementName: The *string* specifying the root element name for which the tree structure should be generated.
        - _`extractedAttributes`: The *dictionary* with information and descriptions data about attributes.
          See details in :ref:`ExtractService`.
        - _`extractedElements`: The *dictionary* with information and descriptions data about elements.
          See details in :ref:`ExtractService`.
        - _`extractedTypes`: The *dictionary* with information and descriptions data about types.
          See details in :ref:`ExtractService`.
        - _`surroundingChars`: The *list* of tuples with the pairs of chars, which can surround attribute, element and type
          names in descriptions. It is used for marking the occurrences of attribute, element and type names to generate
          cross-references in descriptions. It has the following structure: ``LIST[TUPLE(2*STRING)]``.
        - standardStringParts: The *dictionary* with standard string parts used to generate the link to the standard for each
          attribute, element and type. It has the following structure: ``DICTIONARY{STRING:TUPLE(2*STRING)}``.
        - outputStandard: The *boolean* specifying whether the link to the standard should be generated for each attribute, element
          and type. Note that the standard is either the official documentation or the schema in cases when the documentation is not
          available in such a detailed form or uses some other data format than HTML or XHTML. Use true (True) only when
          the given XSD schema is supported, otherwise use false (False) or add the support for the XSD schema by providing
          corresponding standard string parts.
        '''
        if outputStandard: standardFixedPart, standardVariablePart = standardStringParts[self.refName + '--' + self.refVersion]

        attributeNames, elementNames, typeNames = extractedAttributes.keys(), extractedElements.keys(), extractedTypes.keys()
        attributeNames.sort(key=unicode.lower)
        elementNames.sort(key=unicode.lower)
        typeNames.sort(key=unicode.lower)

        with open('Ref--' + self.refName + '--' + self.refVersion + '--AllExtractedInfoAndDescription--'
                  + self.currentVersion + '.xhtml', 'w') as f:
            f.write('<!DOCTYPE html>\n\n')
            f.write('<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">\n\n')
            f.write('<head>\n')
            f.write('<meta charset="utf-8"/>\n')
            f.write('<link rel="stylesheet" type="text/css" href="./Theme/Style.css"/>\n')
            f.write('<link rel="shortcut icon" type="image/x-icon" href="./Theme/Favicon.png"/>\n')
            f.write('<title>' + self.refName + ' ' + self.refVersion + ' Reference</title>\n')
            f.write('</head>\n\n')
            f.write('<body>\n\n')
            f.write('<div class="header"><h1>' + self.refName + ' ' + self.refVersion + ' Reference</h1></div>\n\n')
            f.write('<div class="menu">\n')

            if len(self.refName) > 14: f.write('<h3>' + self.refName[:12] + '... ' + self.refVersion + ' Menu</h3>\n')
            else: f.write('<h3>' + self.refName + ' ' + self.refVersion + ' Menu</h3>\n')

            f.write('<ul>\n')

            if rootElementName: f.write('<li><a href="#tree-structure">Tree Structure</a></li>\n')

            f.write(self._outputMenuStructure(attributeNames, 'att'))
            f.write(self._outputMenuStructure(elementNames, 'el'))
            f.write(self._outputMenuStructure(typeNames, 't'))
            f.write('</ul>\n')
            f.write('</div>\n\n')
            f.write('<div class="content">\n\n')

            if rootElementName:
                processedElements = [rootElementName]
                f.write('<h2 class="separator">OVERVIEW</h2>\n\n')
                f.write('<h2 id="tree-structure">Tree Structure</h2>\n\n')
                f.write('<div class="subcontent tree-structure">\n')
                f.write('<input id="tg-i-tsatt" type="checkbox"/>\n')
                f.write('<label for="tg-i-tsatt">Toggle Attributes</label>\n')
                f.write('<ul>\n')
                f.write('<li>Root element: <a href="#el-' + rootElementName + '">' + rootElementName + '</a>')

                if 'attributes' in extractedElements[rootElementName]:
                    rootElementAttributes = extractedElements[rootElementName]['attributes']
                    rootElementAttributes.sort()
                else: rootElementAttributes = None

                if rootElementAttributes: f.write('<span class="tg-i-tsatt"> (Attributes: '
                                                  + ', '.join('<a href="#att-' + i + '">' + i + '</a>'
                                                              for i in rootElementAttributes) + ')</span>\n')
                else: f.write('\n')

                self._outputTreeStructure(rootElementName, extractedElements, processedElements, f)
                f.write('</ul>\n')
                f.write('</div>\n\n')

            f.write('<h2 class="separator">DETAILS</h2>\n\n')

            if len(attributeNames) > 0:
                with open('Ref--' + self.refName + '--' + self.refVersion + '--Attributes--ExtractedInfo--'
                          + self.currentVersion + '.xml', 'w') as fInfo:
                    with open('Ref--' + self.refName + '--' + self.refVersion + '--Attributes--ExtractedDescription--'
                              + self.currentVersion + '.xml', 'w') as fDescription:
                        f.write('<h2 id="attributes">Attributes</h2>\n\n')
                        fInfo.write('<?xml version="1.0" encoding="utf-8"?>\n\n<attributes mat="Material(\'Ref--'
                                    + self.refName + '--' + self.refVersion + '\', \'science\', \''
                                    + self.refName + ' ' + self.refVersion + ' Reference\')">\n\n')
                        fDescription.write('<?xml version="1.0" encoding="utf-8"?>\n\n<attributes mat="Material(\'Ref--'
                                           + self.refName + '--' + self.refVersion + '\', \'science\', \''
                                           + self.refName + ' ' + self.refVersion + ' Reference\')">\n\n')

                        for aName in attributeNames:
                            aParents, aDescriptionXml, aDescriptionXhtml = self._outputAttributeData(aName, extractedAttributes,
                                                                                                     attributeNames, elementNames,
                                                                                                     surroundingChars)

                            if not outputStandard: standard = ''
                            else: standard = standardFixedPart + standardVariablePart.replace('(N)', aName).replace('(T)',
                                                                                                                    'attribute')

                            # Output the attribute info and description into the XHTML file.
                            f.write('<div class="subcontent attribute">\n')
                            f.write('<h3 id="att-' + aName + '">' + aName + '</h3>\n')
                            f.write('<ul>\n')

                            if outputStandard: f.write('<li>Standard: <a href="' + standard + '">' + standard + '</a></li>\n')

                            f.write('<li>Type: attribute</li>\n')

                            if aParents: f.write('<li>Parents: ' + ', '.join('<a href="#el-' + i + '">' + i + '</a>'
                                                                             for i in aParents) + '</li>\n')
                            else: f.write('<li>Parents: NONE</li>\n')

                            f.write('</ul>\n')

                            if aDescriptionXhtml:
                                f.write('<h4>Description:</h4>\n')
                                f.write('<div class="description">\n')
                                f.write(aDescriptionXhtml + '\n')
                                f.write('</div>\n')

                            f.write('</div>\n\n')

                            # Output the attribute info into the XML file.
                            fInfo.write('<attribute name="' + aName + '" standard="' + standard + '">\n')

                            if aParents: fInfo.write('<parents>' + ','.join(i for i in aParents) + '</parents>\n')
                            else: fInfo.write('<parents>NONE</parents>\n')

                            fInfo.write('</attribute>\n\n')

                            # Output the attribute description into the XML file.
                            fDescription.write('<attribute name="' + aName + '" standard="' + standard + '">\n')
                            fDescription.write('<concept/>\n')

                            if aDescriptionXml: fDescription.write('<description>\n' + aDescriptionXml + '\n</description>\n')
                            else:
                                fDescription.write('<description/>\n')
                                self.printMessage(message=str('INFO: Found no description for attribute "' + aName + '".'))

                            fDescription.write('</attribute>\n\n')

                        fInfo.write('</attributes>')
                        fDescription.write('</attributes>')

            if len(elementNames) > 0:
                with open('Ref--' + self.refName + '--' + self.refVersion + '--Elements--ExtractedInfo--'
                          + self.currentVersion + '.xml', 'w') as fInfo:
                    with open('Ref--' + self.refName + '--' + self.refVersion + '--Elements--ExtractedDescription--'
                              + self.currentVersion + '.xml', 'w') as fDescription:
                        f.write('<h2 id="elements">Elements</h2>\n\n')
                        fInfo.write('<?xml version="1.0" encoding="utf-8"?>\n\n<elements mat="Material(\'Ref--'
                                    + self.refName + '--' + self.refVersion + '\', \'science\', \''
                                    + self.refName + ' ' + self.refVersion + ' Reference\')">\n\n')
                        fDescription.write('<?xml version="1.0" encoding="utf-8"?>\n\n<elements mat="Material(\'Ref--'
                                           + self.refName + '--' + self.refVersion + '\', \'science\', \''
                                           + self.refName + ' ' + self.refVersion + ' Reference\')">\n\n')

                        for eName in elementNames:
                            eAttributes, eChildren, eParents, eDescriptionXml, eDescriptionXhtml = self._outputElementData(eName,
                                extractedElements, attributeNames, elementNames, surroundingChars)

                            if not outputStandard: standard = ''
                            else: standard = standardFixedPart + standardVariablePart.replace('(N)', eName).replace('(T)',
                                                                                                                    'element')

                            # Output the element info and description into the XHTML file.
                            f.write('<div class="subcontent element">\n')
                            f.write('<h3 id="el-' + eName + '">' + eName + '</h3>\n')
                            f.write('<ul>\n')

                            if outputStandard: f.write('<li>Standard: <a href="' + standard + '">' + standard + '</a></li>\n')

                            f.write('<li>Type: element</li>\n')

                            if eAttributes: f.write('<li>Attributes: ' + ', '.join('<a href="#att-' + i + '">' + i + '</a>'
                                                                                   for i in eAttributes) + '</li>\n')
                            else: f.write('<li>Attributes: NONE</li>\n')

                            if eChildren: f.write('<li>Children: ' + ', '.join('<a href="#el-' + i + '">' + i + '</a>'
                                                                               for i in eChildren) + '</li>\n')
                            else: f.write('<li>Children: NONE</li>\n')

                            if eParents: f.write('<li>Parents: ' + ', '.join('<a href="#el-' + i + '">' + i + '</a>'
                                                                             for i in eParents) + '</li>\n')
                            else: f.write('<li>Parents: NONE</li>\n')

                            f.write('</ul>\n')

                            if eDescriptionXhtml:
                                f.write('<h4>Description:</h4>\n')
                                f.write('<div class="description">\n')
                                f.write(eDescriptionXhtml + '\n')
                                f.write('</div>\n')

                            f.write('</div>\n\n')

                            # Output the element info into the XML file.
                            fInfo.write('<element name="' + eName + '" standard="' + standard + '">\n')

                            if eAttributes: fInfo.write('<attributes>' + ','.join(i for i in eAttributes) + '</attributes>\n')
                            else: fInfo.write('<attributes/>\n')

                            if eChildren: fInfo.write('<children>' + ','.join(i for i in eChildren) + '</children>\n')
                            else: fInfo.write('<children/>\n')

                            if eParents: fInfo.write('<parents>' + ','.join(i for i in eParents) + '</parents>\n')
                            else: fInfo.write('<parents/>\n')

                            fInfo.write('</element>\n\n')

                            # Output the element description into the XML file.
                            fDescription.write('<element name="' + eName + '" standard="' + standard + '">\n')
                            fDescription.write('<concept/>\n')

                            if eDescriptionXml: fDescription.write('<description>\n' + eDescriptionXml + '\n</description>\n')
                            else:
                                fDescription.write('<description/>\n')
                                self.printMessage(message=str('INFO: Found no description for element "' + eName + '".'))

                            fDescription.write('</element>\n\n')

                        fInfo.write('</elements>')
                        fDescription.write('</elements>')

            if len(typeNames) > 0:
                with open('Ref--' + self.refName + '--' + self.refVersion + '--Types--ExtractedInfo--'
                          + self.currentVersion + '.xml', 'w') as fInfo:
                    with open('Ref--' + self.refName + '--' + self.refVersion + '--Types--ExtractedDescription--'
                              + self.currentVersion + '.xml', 'w') as fDescription:
                        f.write('<h2 id="types">Types</h2>\n\n')
                        fInfo.write('<?xml version="1.0" encoding="utf-8"?>\n\n<types mat="Material(\'Ref--'
                                    + self.refName + '--' + self.refVersion + '\', \'science\', \''
                                    + self.refName + ' ' + self.refVersion + ' Reference\')">\n\n')
                        fDescription.write('<?xml version="1.0" encoding="utf-8"?>\n\n<types mat="Material(\'Ref--'
                                           + self.refName + '--' + self.refVersion + '\', \'science\', \''
                                           + self.refName + ' ' + self.refVersion + ' Reference\')">\n\n')

                        for tName in typeNames:
                            (tForm, tAttributes, tChildren, tUsagesAttributes, tUsagesElements,
                             tDescriptionXml, tDescriptionXhtml) = self._outputTypeData(tName, extractedTypes,
                                attributeNames, elementNames, surroundingChars)
                            tName = tName[2:]

                            if not outputStandard: standard = ''
                            else: standard = standardFixedPart + standardVariablePart.replace('(N)', tName).replace('(T)', tForm)

                            # Output the type info and description into the XHTML file.
                            f.write('<div class="subcontent type">\n')
                            f.write('<h3 id="t-' + tName + '">' + tName + '</h3>\n')
                            f.write('<ul>\n')

                            if outputStandard: f.write('<li>Standard: <a href="' + standard + '">' + standard + '</a></li>\n')

                            f.write('<li>Type: ' + tForm + '</li>\n')

                            if tAttributes: f.write('<li>Attributes: ' + ', '.join('<a href="#att-' + i + '">' + i + '</a>'
                                                                                   for i in tAttributes) + '</li>\n')
                            else: f.write('<li>Attributes: NONE</li>\n')

                            if tChildren: f.write('<li>Children: ' + ', '.join('<a href="#el-' + i + '">' + i + '</a>'
                                                                               for i in tChildren) + '</li>\n')
                            else: f.write('<li>Children: NONE</li>\n')

                            if tUsagesAttributes: f.write('<li>Usages in attributes: ' + ', '.join('<a href="#att-' + i + '">'
                                + i + '</a>' for i in tUsagesAttributes) + '</li>\n')
                            else: f.write('<li>Usages in attributes: NONE</li>\n')

                            if tUsagesElements: f.write('<li>Usages in elements: ' + ', '.join('<a href="#el-' + i + '">'
                                + i + '</a>' for i in tUsagesElements) + '</li>\n')
                            else: f.write('<li>Usages in elements: NONE</li>\n')

                            f.write('</ul>\n')

                            if tDescriptionXhtml:
                                f.write('<h4>Description:</h4>\n')
                                f.write('<div class="description">\n')
                                f.write(tDescriptionXhtml + '\n')
                                f.write('</div>\n')

                            f.write('</div>\n\n')

                            # Output the type info into the XML file.
                            fInfo.write('<type form="' + tForm + '" name="' + tName + '" standard="' + standard + '">\n')

                            if tAttributes: fInfo.write('<attributes>' + ','.join(i for i in tAttributes) + '</attributes>\n')
                            else: fInfo.write('<attributes/>\n')

                            if tChildren: fInfo.write('<children>' + ','.join(i for i in tChildren) + '</children>\n')
                            else: fInfo.write('<children/>\n')

                            fInfo.write('<usages>\n')

                            if tUsagesAttributes: fInfo.write('<attributes>' + ','.join(i for i in tUsagesAttributes)
                                                              + '</attributes>\n')
                            else: fInfo.write('<attributes/>\n')

                            if tUsagesElements: fInfo.write('<elements>' + ','.join(i for i in tUsagesElements) + '</elements>\n')
                            else: fInfo.write('<elements/>\n')

                            fInfo.write('</usages>\n')
                            fInfo.write('</type>\n\n')

                            # Output the type description into the XML file.
                            fDescription.write('<type form="' + tForm + '" name="' + tName + '" standard="' + standard + '">\n')
                            fDescription.write('<concept/>\n')

                            if tDescriptionXml: fDescription.write('<description>\n' + tDescriptionXml + '\n</description>\n')
                            else:
                                fDescription.write('<description/>\n')
                                self.printMessage(message=str('INFO: Found no description for type "' + tName + '".'))

                            fDescription.write('</type>\n\n')

                        fInfo.write('</types>')
                        fDescription.write('</types>')

            f.write('</div>\n\n')
            f.write('<div class="footer">Generated using DATA FORMATS ANALYZER 1-1-1 on ' + datetime.date.today().isoformat()
                    + '.</div>\n\n')
            f.write('</body>\n\n')
            f.write('</html>')

        self.printMessage(message='INFO: All attributes, elements and types saved in XHTML and XML files.')

    def _outputMenuStructure(self, extractedNames, namesType):
        '''
        Outputs a generated menu structure for the given extracted names.

        **Parameters:**

        - extractedNames: The *list* of extracted names to be processed. It has the following structure: ``LIST[STRING]``.
        - namesType: The *string* specifying the type of extracted names.

        **Returns:** The *string* with the generated menu structure.
        '''
        if len(extractedNames) > 0:
            if namesType == 'att': namesTitle = 'attributes'
            elif namesType == 'el': namesTitle = 'elements'
            elif namesType == 't':
                namesTitle = 'types'
                extractedNames = [name[2:] for name in extractedNames]
            else: namesTitle = namesType

            menuStructure = ('<li>\n<a href="#' + namesTitle + '">' + namesTitle.title() + '</a>\n<input id="tg-b-m' + namesType
                             + '" type="checkbox"/>\n<label for="tg-b-m' + namesType + '">Toggle ' + str(len(extractedNames))
                             + ' Items</label>\n<div class="tg-b-m' + namesType + ' submenu">\n<ul>\n')
            processedNames = {}

            for name in extractedNames:
                nameInitial = name[0].upper()

                if nameInitial not in processedNames: processedNames[nameInitial] = [name, 1]
                else: processedNames[nameInitial][1] += 1

            processedNamesInitials = processedNames.keys()
            processedNamesInitials.sort()
            menuStructure += '\n'.join('<li><a href="#' + namesType + '-' + processedNames[i][0] + '">' + i + '</a> ('
                                   + str(processedNames[i][1]) + ' items)</li>' for i in processedNamesInitials)
            return (menuStructure + '\n</ul>\n</div>\n</li>\n').replace('(1 items)', '(1 item)')
        else: return ''

    def _outputTreeStructure(self, elementName, extractedElements, processedElements, f):
        '''
        Outputs a generated tree structure for the given element. Recursive approach is used to process all elements in
        extractedElements_ *dictionary*.

        **Parameters:**

        - elementName: The *string* with the current element name, which will be used to get information and descriptions data.
        - extractedElements_: The *dictionary* with information and descriptions data about elements.
        - processedElements: The *list* of all element names, which were processed so far. This is used to prevent unnecessary
          recursion. It has the following structure: ``LIST[STRING]``.
        - f: The *file* where the output is written.
        '''
        if 'children' in extractedElements[elementName]:
            eChildren = extractedElements[elementName]['children']
            eChildren.sort()
        else: eChildren = None

        if eChildren:
            processedElements += eChildren
            f.write('<ul>\n')

            for eName in eChildren:
                if 'attributes' in extractedElements[eName]:
                    eAttributes = extractedElements[eName]['attributes']
                    eAttributes.sort()
                else: eAttributes = None

                if (eName in processedElements) and ('children' in extractedElements[eName]):
                    processed = True

                    for eChild in extractedElements[eName]['children']:
                        if eChild not in processedElements:
                            processed = False
                            f.write('<li>Element: <a href="#el-' + eName + '">' + eName + '</a>')

                            if eAttributes: f.write('<span class="tg-i-tsatt"> (Attributes: '
                                                    + ', '.join('<a href="#att-' + i + '">' + i + '</a>'
                                                                for i in eAttributes) + ')</span>\n')
                            else: f.write('\n')

                            self._outputTreeStructure(eName, extractedElements, processedElements, f)
                            break

                    if processed:
                        f.write('<li>Element: <a href="#el-' + eName + '">' + eName + '</a>')

                        if eAttributes: f.write('<span class="tg-i-tsatt"> (Attributes: '
                                                + ', '.join('<a href="#att-' + i + '">' + i + '</a>'
                                                            for i in eAttributes) + ')</span> ...</li>\n')
                        else: f.write(' ...</li>\n')
                else:
                    f.write('<li>Element: <a href="#el-' + eName + '">' + eName + '</a>')

                    if eAttributes: f.write('<span class="tg-i-tsatt"> (Attributes: '
                                            + ', '.join('<a href="#att-' + i + '">' + i + '</a>'
                                                        for i in eAttributes) + ')</span></li>\n')
                    else: f.write('</li>\n')

            f.write('</ul>\n</li>\n')

    def _outputAttributeData(self, attributeName, extractedAttributes, attributeNames, elementNames, surroundingChars):
        '''
        Returns information and descriptions data about the given attribute name from extractedAttributes_ *dictionary*.

        **Parameters:**

        - attributeName: The *string* with the current attribute name, which will be used to get information and descriptions data.
        - extractedAttributes_: The *dictionary* with information and descriptions data about attributes.
        - attributeNames: The *list* of all attribute names, which will be used to mark the occurrences of other
          attribute names in the descriptions data of the current attribute. It has the following structure: ``LIST[STRING]``.
        - elementNames: The *list* of all element names, which will be used to mark the occurrences of element names
          in the descriptions data of the current attribute. It has the following structure: ``LIST[STRING]``.
        - surroundingChars_: The *list* of tuples with the pairs of chars, which can surround the attribute and element
          names in descriptions.

        **Returns:** The *tuple* with information and descriptions data about the given attribute name. It has the following
        structure::

            TUPLE(
                LIST[Parent name (STRING)],
                Attribute XML documentation (STRING),
                Attribute XHTML documentation (STRING)
            )
        '''
        if 'parents' in extractedAttributes[attributeName]:
            aParents = extractedAttributes[attributeName]['parents']
            aParents.sort()
        else: aParents = None

        if 'description' in extractedAttributes[attributeName]:
            aDescriptionXml = aDescriptionXhtml = '\n'.join(
                descriptionPart for descriptionPart in extractedAttributes[attributeName]['description'])

            # Mark the occurrences of element names in the descriptions data of the current attribute.
            for name in elementNames:
                for chL, chR in surroundingChars:
                    aDescriptionXml = aDescriptionXml.replace(chL + name + chR, chL + '<el>' + name + '</el>' + chR)
                    aDescriptionXhtml = aDescriptionXhtml.replace(
                        chL + name + chR, chL + '<a href="#el-' + name + '">' + name + '</a>' + chR)

            # Mark the occurrences of other attribute names in the descriptions data of the current attribute.
            for name in attributeNames:
                if name != attributeName:
                    for chL, chR in surroundingChars:
                        aDescriptionXml = aDescriptionXml.replace(chL + name + chR, chL + '<att>' + name + '</att>' + chR)
                        aDescriptionXhtml = aDescriptionXhtml.replace(
                            chL + name + chR, chL + '<a href="#att-' + name + '">' + name + '</a>' + chR)
        else: aDescriptionXml = aDescriptionXhtml = None

        return (aParents, aDescriptionXml, aDescriptionXhtml)

    def _outputElementData(self, elementName, extractedElements, attributeNames, elementNames, surroundingChars):
        '''
        Returns information and descriptions data about the given element name from extractedElements_ *dictionary*.

        **Parameters:**

        - elementName: The *string* with the current element name, which will be used to get information and descriptions data.
        - extractedElements_: The *dictionary* with information and descriptions data about elements.
        - attributeNames: The *list* of all attribute names, which will be used to mark the occurrences of attribute names
          in the descriptions data of the current element. It has the following structure: ``LIST[STRING]``.
        - elementNames: The *list* of all element names, which will be used to mark the occurrences of other
          element names in the descriptions data of the current element. It has the following structure: ``LIST[STRING]``.
        - surroundingChars_: The *list* of tuples with the pairs of chars, which can surround the attribute and element
          names in descriptions.

        **Returns:** The *tuple* with information and descriptions data about the given element name. It has the following
        structure::

            TUPLE(
                LIST[Attribute name (STRING)],
                LIST[Child name (STRING)],
                LIST[Parent name (STRING)],
                Element XML documentation (STRING),
                Element XHTML documentation (STRING)
            )
        '''
        if 'attributes' in extractedElements[elementName]:
            eAttributes = extractedElements[elementName]['attributes']
            eAttributes.sort()
        else: eAttributes = None

        if 'children' in extractedElements[elementName]:
            eChildren = extractedElements[elementName]['children']
            eChildren.sort()
        else: eChildren = None

        if 'parents' in extractedElements[elementName]:
            eParents = extractedElements[elementName]['parents']
            eParents.sort()
        else: eParents = None

        if 'description' in extractedElements[elementName]:
            eDescriptionXml = eDescriptionXhtml = '\n'.join(
                descriptionPart for descriptionPart in extractedElements[elementName]['description'])

            # Mark the occurrences of other element names in the descriptions data of the current element.
            for name in elementNames:
                if name != elementName:
                    for chL, chR in surroundingChars:
                        eDescriptionXml = eDescriptionXml.replace(chL + name + chR, chL + '<el>' + name + '</el>' + chR)
                        eDescriptionXhtml = eDescriptionXhtml.replace(
                            chL + name + chR, chL + '<a href="#el-' + name + '">' + name + '</a>' + chR)

            # Mark the occurrences of attribute names in the descriptions data of the current element.
            for name in attributeNames:
                for chL, chR in surroundingChars:
                    eDescriptionXml = eDescriptionXml.replace(chL + name + chR, chL + '<att>' + name + '</att>' + chR)
                    eDescriptionXhtml = eDescriptionXhtml.replace(
                        chL + name + chR, chL + '<a href="#att-' + name + '">' + name + '</a>' + chR)
        else: eDescriptionXml = eDescriptionXhtml = None

        return (eAttributes, eChildren, eParents, eDescriptionXml, eDescriptionXhtml)

    def _outputTypeData(self, typeName, extractedTypes, attributeNames, elementNames, surroundingChars):
        '''
        Returns information and descriptions data about the given type name from extractedTypes_ *dictionary*.

        **Parameters:**

        - typeName: The *string* with the current type name, which will be used to get information and descriptions data.
        - extractedTypes_: The *dictionary* with information and descriptions data about types.
        - attributeNames: The *list* of all attribute names, which will be used to mark the occurrences of attribute names
          in the descriptions data of the current type. It has the following structure: ``LIST[STRING]``.
        - elementNames: The *list* of all element names, which will be used to mark the occurrences of element names
          in the descriptions data of the current type. It has the following structure: ``LIST[STRING]``.
        - surroundingChars_: The *list* of tuples with the pairs of chars, which can surround the attribute and element
          names in descriptions.

        **Returns:** The *tuple* with information and descriptions data about the given type name. It has the following
        structure::

            TUPLE(
                Type form (STRING),
                LIST[Attribute name (STRING)],
                LIST[Child name (STRING)],
                LIST[Usage in attribute name (STRING)],
                LIST[Usage in element name (STRING)],
                Type XML documentation (STRING),
                Type XHTML documentation (STRING)
            )
        '''
        if typeName.startswith('C-'): tForm = 'complexType'
        else: tForm = 'simpleType'

        if 'attributes' in extractedTypes[typeName]:
            tAttributes = extractedTypes[typeName]['attributes']
            tAttributes.sort()
        else: tAttributes = None

        if 'children' in extractedTypes[typeName]:
            tChildren = extractedTypes[typeName]['children']
            tChildren.sort()
        else: tChildren = None

        if 'usages' in extractedTypes[typeName]:
            if 'attributes' in extractedTypes[typeName]['usages']:
                tUsagesAttributes = extractedTypes[typeName]['usages']['attributes']
                tUsagesAttributes.sort()
            else: tUsagesAttributes = None

            if 'elements' in extractedTypes[typeName]['usages']:
                tUsagesElements = extractedTypes[typeName]['usages']['elements']
                tUsagesElements.sort()
            else: tUsagesElements = None
        else: tUsagesAttributes = tUsagesElements = None

        if 'description' in extractedTypes[typeName]:
            tDescriptionXml = tDescriptionXhtml = '\n'.join(
                descriptionPart for descriptionPart in extractedTypes[typeName]['description'])

            # Mark the occurrences of element names in the descriptions data of the current type.
            for name in elementNames:
                for chL, chR in surroundingChars:
                    tDescriptionXml = tDescriptionXml.replace(chL + name + chR, chL + '<el>' + name + '</el>' + chR)
                    tDescriptionXhtml = tDescriptionXhtml.replace(
                        chL + name + chR, chL + '<a href="#el-' + name + '">' + name + '</a>' + chR)

            # Mark the occurrences of attribute names in the descriptions data of the current type.
            for name in attributeNames:
                for chL, chR in surroundingChars:
                    tDescriptionXml = tDescriptionXml.replace(chL + name + chR, chL + '<att>' + name + '</att>' + chR)
                    tDescriptionXhtml = tDescriptionXhtml.replace(
                        chL + name + chR, chL + '<a href="#att-' + name + '">' + name + '</a>' + chR)
        else: tDescriptionXml = tDescriptionXhtml = None

        return (tForm, tAttributes, tChildren, tUsagesAttributes, tUsagesElements, tDescriptionXml, tDescriptionXhtml)
